#include "acctcl.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <netdb.h>

extern int inet_aton (const char *cp, struct in_addr *addr) ;

struct ipfilter
{
    struct line *items ;
    int nitems ;
} ;

struct line				/* one element of a filter */
{
    int    catch ;			/* 1 if must catch, 0 if reject */
    struct in_addr mask ;		/* mask to apply to the given address */
    struct in_addr ref ;		/* reference address to compare to */
} ;

/* commands */
int Ipfilter_Create (ClientData, Tcl_Interp *, int, char *[]) ;
int Ipfilter_Test   (ClientData, Tcl_Interp *, int, char *[]) ;
int Ipfilter_Name   (ClientData, Tcl_Interp *, int, char *[]) ;

/* sub-commands */
int Ipfilter_Set    (ClientData, Tcl_Interp *, int, char *[]) ;
int Ipfilter_Delete (ClientData, Tcl_Interp *, int, char *[]) ;
int Ipfilter_Match  (ClientData, Tcl_Interp *, int, char *[]) ;

/******************************************************************************
Driver procedure
******************************************************************************/

int Ipfilter_Cmd (ClientData cd, Tcl_Interp *interp,
						int argc, char *argv [])
{
    static optable op [] =
    {
	{ "test",   Ipfilter_Test,   },
	{ "name",   Ipfilter_Name,   },
	{ "create", Ipfilter_Create, },
    } ;

    return callproc (op, NTAB (op), cd, interp, argc, argv) ;
}

/******************************************************************************
Utility functions
******************************************************************************/

int ipf_parse (Tcl_Interp *interp, char *str, struct ipfilter *ipf)
{
    int largc ;
    char **largv ;
    int i ;
    struct line *items ;

    if (Tcl_SplitList (interp, str, &largc, &largv) == TCL_ERROR)
	return TCL_ERROR ;

    if (largc % 3 != 0)
    {
	free ((char *) largv) ;
	interp->result = "invalid filter (length should be a multiple of 3)" ;
	return TCL_ERROR ;
    }

    largc /= 3 ;

    items = (struct line *) calloc (largc, sizeof (struct line)) ;
    if (items == NULL)
    {
	free ((char *) largv) ;
	interp->result = "cannot allocate memory" ;
	return TCL_ERROR ;
    }

    for (i = 0 ; i < largc ; i++)
    {
	struct line *f ;
	char **tab ;

	f = & items [i] ;
	tab = & largv [3 * i] ;

	if (EQ (tab [0], "+"))
	    f->catch = 1 ;
	else if (EQ (tab [0], "-"))
	    f->catch = 0 ;
	else
	{
	    free ((char *) largv) ;
	    free ((char *) items) ;
	    interp->result = "invalid filter (should contain '+' or '-')" ;
	    return TCL_ERROR ;
	}

	if (inet_aton (tab [1], &f->ref) == 0)
	{
	    free ((char *) largv) ;
	    free ((char *) items) ;
	    interp->result = "invalid IP address" ;
	    return TCL_ERROR ;
	}

	if (inet_aton (tab [2], &f->mask) == 0)
	{
	    free ((char *) largv) ;
	    free ((char *) items) ;
	    interp->result = "invalid IP mask" ;
	    return TCL_ERROR ;
	}
	f->mask.s_addr = ~ f->mask.s_addr ;
    }

    ipf->nitems = largc ;
    ipf->items = items ;

    free ((char *) largv) ;

    return TCL_OK ;
}

/******************************************************************************
Command "ipfilter test"
******************************************************************************/

int Ipfilter_Test (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct in_addr a ;

    if (argc != 3)
    {
	interp->result = "wrong # args: should be ipfilter test ipaddr" ;
	return TCL_ERROR ;
    }

    interp->result = inet_aton (argv [2], &a) ? "1" : "0" ;
    return TCL_OK ;
}


/******************************************************************************
Command "ipfilter name"
******************************************************************************/

int Ipfilter_Name (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct in_addr a ;
    struct hostent *hp ;

    if (argc != 3)
    {
	interp->result = "wrong # args: should be ipfilter name ipaddr" ;
	return TCL_ERROR ;
    }

    if (inet_aton (argv [2], &a) == 0)
    {
	interp->result = "invalid ip address" ;
	return TCL_ERROR ;
    }

    hp = gethostbyaddr ((char *) &a, sizeof a, AF_INET) ;
    if (hp == NULL)
	interp->result = "" ;
    else
	Tcl_SetResult (interp, hp->h_name, TCL_VOLATILE) ;

    return TCL_OK ;
}


/******************************************************************************
Command "ipfilter create"
******************************************************************************/

int Filter_SubCmd (ClientData, Tcl_Interp *, int, char *[]) ;
void Filter_Delete (ClientData) ;

int Ipfilter_Create (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct ipfilter *ipf ;

    if (argc != 3)
    {
	interp->result = "wrong # args: should be ipfilter create name" ;
	return TCL_ERROR ;
    }

    ipf = (struct ipfilter *) malloc (sizeof (struct ipfilter)) ;
    if (ipf == NULL)
    {
	interp->result = "not enough memory" ;
	return TCL_ERROR ;
    }
    ipf->items = NULL ;
    ipf->nitems = 0 ;

    Tcl_CreateCommand (interp, argv [2],
		Filter_SubCmd, (ClientData) ipf, Filter_Delete) ;

    interp->result = "" ;
    return TCL_OK ;
}

void Filter_Delete (ClientData cd)
{
    struct ipfilter *ipf ;

    ipf = (struct ipfilter *) cd ;
    if (ipf->nitems != 0)
	free ((char *) ipf->items) ;
    free ((char *) ipf) ;
}

int Filter_SubCmd (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    static optable op [] =
    {
	{ "set",    Ipfilter_Set,    },
	{ "delete", Ipfilter_Delete, },
	{ "match",  Ipfilter_Match,  },
    } ;

    return callproc (op, NTAB (op), cd, interp, argc, argv) ;
}

/******************************************************************************
Command "<ipfhandle> set"
******************************************************************************/

int Ipfilter_Set (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct ipfilter *ipf ;

    if (argc != 3)
    {
	interp->result = "wrong # args: should be ipfilter set list" ;
	return TCL_ERROR ;
    }

    /*
     * Parse the filter and allocate memory for it
     */

    ipf = (struct ipfilter *) cd ;
    if (ipf->nitems != 0)
	free ((char *) ipf->items) ;

    if (ipf_parse (interp, argv [2], ipf) == TCL_ERROR)
	return TCL_ERROR ;

    /*
     * Return the new filter index
     */

    interp->result = "" ;
    return TCL_OK ;
}


/******************************************************************************
Command "<ipfhandle> delete"
******************************************************************************/

int Ipfilter_Delete (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    if (argc != 2)
    {
	interp->result = "wrong # args: should be ipfilter delete" ;
	return TCL_ERROR ;
    }

    Tcl_DeleteCommand (interp, argv [0]) ;

    interp->result = "" ;
    return TCL_OK ;
}

/******************************************************************************
Command "<ipfhandle> match"
******************************************************************************/

int Ipfilter_Match (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct in_addr a ;
    int i ;
    char *result ;
    struct ipfilter *ipf ;

    if (argc != 3)
    {
	interp->result = "wrong # args: should be ipfilter match address" ;
	return TCL_ERROR ;
    }

    ipf = (struct ipfilter *) cd ;

    if (inet_aton (argv [2], &a) == 0)
    {
	interp->result = "invalid ip address" ;
	return TCL_ERROR ;
    }

    result = NULL ;
    i = 0 ;
    while (i < ipf->nitems && result == NULL)
    {
	if ((a.s_addr & ipf->items [i].mask.s_addr) == ipf->items [i].ref.s_addr)
	    result = ipf->items [i].catch ? "1" : "0" ;
	i++ ;
    }
    if (result == NULL)
	result = "0" ;

    interp->result = result ;
    return TCL_OK ;
}
