#include "acctcl.h"

#include <netinet/in.h>
#include <arpa/inet.h>

extern int get64 (Tcl_Interp *interp, char *str, int64 *pn) ;
extern int print64 (Tcl_Interp *interp, char *str, int64 n, int base) ;

extern int inet_aton (const char *cp, struct in_addr *addr) ;

#define	DEFAULT_HASH_SIZE	1789
struct traffic
{
    int hashsize ;
    int nelem ;
    struct pair **hashtable ;
} ;

struct pair
{
    struct in_addr ip1 ;
    struct in_addr ip2 ;
    int64 traf12 ;			/* nb bytes 1 -> 2 */
    int64 traf21 ;			/* nb bytes 2 -> 1 */
    struct pair *next ;
} ;

/* traffic commands */
int Traffic_Create (ClientData, Tcl_Interp *, int, char *[]) ;

/* traffic subcommands */
int Subtraffic_Add    (ClientData, Tcl_Interp *, int, char *[]) ;
int Subtraffic_Get    (ClientData, Tcl_Interp *, int, char *[]) ;
int Subtraffic_Pairs  (ClientData, Tcl_Interp *, int, char *[]) ;
int Subtraffic_Sort   (ClientData, Tcl_Interp *, int, char *[]) ;
int Subtraffic_Sortip (ClientData, Tcl_Interp *, int, char *[]) ;
int Subtraffic_Total  (ClientData, Tcl_Interp *, int, char *[]) ;
int Subtraffic_Stats  (ClientData, Tcl_Interp *, int, char *[]) ;
int Subtraffic_Delete (ClientData, Tcl_Interp *, int, char *[]) ;

/******************************************************************************
Driver procedure
******************************************************************************/

int Traffic_Cmd (ClientData cd, Tcl_Interp *interp,
						int argc, char *argv [])
{
    static optable op [] =
    {
	{ "create", Traffic_Create},
    } ;

    return callproc (op, NTAB (op), cd, interp, argc, argv) ;
}

/******************************************************************************
Utility functions
******************************************************************************/

/*
 * Compute a hash value given two IP numbers.
 * The hash value is computed by taking the 3 or 4 low order bits of each
 * byte of each IP address
 * Then, this value is clipped to the hashtable size.
 */

int hash (struct traffic *tf, unsigned long int ip1, unsigned long int ip2)
{
    unsigned int h ;

    h =  (ip1 & 0x0000000f) |
	((ip1 & 0x00000f00) >> (8 - 1*4)) |
	((ip1 & 0x000f0000) >> (16 - 2*4)) |
	((ip1 & 0x07000000) >> (24 - 3*4)) |
	((ip2 & 0x0000000f) << ((4*4-1) - 0)) |
	((ip2 & 0x00000f00) << ((5*4-1) - 8)) |
	((ip2 & 0x000f0000) << ((6*4-1) - 16)) |
	((ip2 & 0x07000000) << ((7*4-1) - 24))
	;
    return h % tf->hashsize ;
}

int need_to_reverse (unsigned long int ip1, unsigned long int ip2)
{
    return ip1 > ip2 ;
}

struct pair *get_pair (int create, struct traffic *tf, unsigned long ip1, unsigned long ip2)
{
    int h ;
    struct pair *p ;

    h = hash (tf, ip1, ip2) ;
    p = tf->hashtable [h] ;

    while (p != NULL && (ip1 != p->ip1.s_addr || ip2 != p->ip2.s_addr))
	p = p->next ;

    if (p == NULL && create)
    {
	p = (struct pair *) malloc (sizeof (struct pair)) ;
	if (p != NULL)
	{
	    p->ip1.s_addr = ip1 ;
	    p->ip2.s_addr = ip2 ;
	    p->traf12 = 0 ;
	    p->traf21 = 0 ;
	    p->next = tf->hashtable [h] ;
	    tf->hashtable [h] = p ;
	    tf->nelem += 1 ;
	}
    }
    return p ;
}

struct pair **build_array (struct traffic *tf)
{
    int nelem ;
    struct pair **elem ;
    struct pair *p ;
    int h, i ;

    nelem = tf->nelem ;
    if (nelem == 0)
	return NULL ;

    elem = (struct pair **) calloc (nelem, sizeof (struct pair *)) ;
    if (elem == NULL)
	return NULL ;

    i = 0 ;
    for (h = 0 ; h < tf->hashsize ; h++)
    {
	p = tf->hashtable [h] ;
	while (p != NULL)
	{
	    elem [i++] = p ;
	    p = p->next ;
	}
    }

    return elem ;
}

void reverse_array (struct pair **elem, int nelem)
{
    int i, j ;
    struct pair *tmp ;

    i = 0 ; j = nelem - 1 ;
    while (i < j)
    {
	tmp = elem [i] ;
	elem [i] = elem [j] ;
	elem [j] = tmp ;

	i++ ;
	j-- ;
    }
}

char *write_array_as_list (struct pair **elem, int nelem)
{
    char *s, *q ;
    struct pair *p ;
    int i ;

    /*
     * each pair needs : "{aaa.bbb.ccc.ddd eee.fff.ggg.hhh} "
     *   - 2 * (4 * 3 + 3) = 2 * 15 bytes = 30 (aaa.bbb.ccc.ddd)
     *   - 2 bytes (braces)
     *   - 1 byte (blanck between 2 address)
     *   - 1 byte (blanck between 2 lists)
     * Do not forget the nul character !
     * Total :
     *   (34 * nbpairs) + 1
     */

    s = (char *) malloc (nelem * 34 + 1) ;
    if (s == NULL)
	return NULL ;

    q = s ;
    for (i = 0 ; i < nelem ; i++)
    {
	p = elem [i] ;
	*q++ = '{' ;
	strcpy (q, inet_ntoa (p->ip1)) ;
	q += strlen (q) ;
	*q++ = ' ' ;
	strcpy (q, inet_ntoa (p->ip2)) ;
	q += strlen (q) ;
	*q++ = '}' ;
	*q++ = ' ' ;
    }
    *q = '\0' ;

    return s ;
}

/******************************************************************************
Command "traffic create"
******************************************************************************/

int Traffic_SubCmd (ClientData, Tcl_Interp *, int, char *[]) ;
void Traffic_Delete (ClientData) ;

int Traffic_Create (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    int hashsize ;
    struct traffic *tf ;

    switch (argc)
    {
	case 3 :
	    hashsize = DEFAULT_HASH_SIZE ;
	    break ;
	case 4 :
	    if (Tcl_GetInt (interp, argv [3], &hashsize) != TCL_OK)
		return TCL_ERROR ;
	    if (hashsize <= 0)
	    {
		interp->result = "hashsize must be > 0" ;
		return TCL_ERROR ;
	    }
	    break ;
	default :
	    interp->result = "wrong # args: should be traffic create name ?hashsize?" ;
	    return TCL_ERROR ;
    }

    tf = (struct traffic *) malloc (sizeof (struct traffic)) ;
    if (tf == NULL)
    {
	interp->result = "not enough memory" ;
	return TCL_ERROR ;
    }
    tf->nelem = 0 ;
    tf->hashsize = hashsize ;
    tf->hashtable = (struct pair **) calloc (hashsize, sizeof (struct pair *)) ;
    if (tf->hashtable == NULL)
    {
	interp->result = "not enough memory" ;
	free ((char *) tf) ;
	return TCL_ERROR ;
    }

    Tcl_CreateCommand (interp, argv [2],
		Traffic_SubCmd, (ClientData) tf, Traffic_Delete) ;

    interp->result = "" ;
    return TCL_OK ;
}

void Traffic_Delete (ClientData cd)
{
    struct traffic *tf ;

    tf = (struct traffic *) cd ;
    if (tf->hashtable != NULL)
    {
	struct pair *cur ;
	int i ;

	for (i = 0 ; i < tf->hashsize ; i++)
	{
	    cur = tf->hashtable [i] ;
	    while (cur != NULL)
	    {
		struct pair *next ;

		next = cur->next ;
		free ((char *) cur) ;
		cur = next ;
	    }
	}
	free ((char *) tf->hashtable) ;
    }
    free ((char *) tf) ;
}

int Traffic_SubCmd (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    static optable op [] =
    {
	{ "add",    Subtraffic_Add,    },
	{ "get",    Subtraffic_Get,    },
	{ "pairs",  Subtraffic_Pairs,  },
	{ "sort",   Subtraffic_Sort,   },
	{ "sortip", Subtraffic_Sortip, },
	{ "total",  Subtraffic_Total,  },
	{ "stats",  Subtraffic_Stats,  },
	{ "delete", Subtraffic_Delete, },
    } ;

    return callproc (op, NTAB (op), cd, interp, argc, argv) ;
}

/******************************************************************************
Command "<traffichandle> add"
******************************************************************************/

int Subtraffic_Add (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct in_addr ip1, ip2 ;
    int64 count ;
    struct traffic *tf ;
    struct pair *p ;
    int reverse ;

    if (argc != 5)
    {
	interp->result = "wrong # args: should be traffic add ip1 ip2 count" ;
	return TCL_ERROR ;
    }

    tf = (struct traffic *) cd ;

    if (inet_aton (argv [2], &ip1) == 0 || inet_aton (argv [3], &ip2) == 0)
    {
	interp->result = "invalid IP address" ;
	return TCL_ERROR ;
    }

    if (get64 (interp, argv [4], &count) == TCL_ERROR)
	return TCL_ERROR ;

    reverse = need_to_reverse (ip1.s_addr, ip2.s_addr) ;

    if (reverse)
	p = get_pair (1, tf, ip1.s_addr, ip2.s_addr) ;
    else
	p = get_pair (1, tf, ip2.s_addr, ip1.s_addr) ;

    if (p == NULL)
    {
	interp->result = "cannot allocate memory" ;
	return TCL_ERROR ;
    }

    p->traf12 += reverse ? 0 : count ;
    p->traf21 += reverse ? count : 0 ;

    interp->result = "" ;
    return TCL_OK ;
}


/******************************************************************************
Command "<traffichandle> get"
******************************************************************************/

int Subtraffic_Get (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct in_addr ip1, ip2 ;
    struct traffic *tf ;
    struct pair *p ;
    int reverse ;

    if (argc != 4)
    {
	interp->result = "wrong # args: should be traffic get ip1 ip2" ;
	return TCL_ERROR ;
    }

    tf = (struct traffic *) cd ;

    if (inet_aton (argv [2], &ip1) == 0 || inet_aton (argv [3], &ip2) == 0)
    {
	interp->result = "invalid IP address" ;
	return TCL_ERROR ;
    }

    reverse = need_to_reverse (ip1.s_addr, ip2.s_addr) ;

    if (reverse)
	p = get_pair (0, tf, ip1.s_addr, ip2.s_addr) ;
    else
	p = get_pair (0, tf, ip2.s_addr, ip1.s_addr) ;

    if (p == NULL)
    {
	interp->result = "" ;
    }
    else
    {
	char str [100], *s ;
	int r ;
	int64 total ;

	s = str ;

	total = p->traf12 + p->traf21 ;
	r = print64 (interp, s, total, 10) ;
	if (r != TCL_OK)
	    return TCL_ERROR ;

	s = s + strlen (s) ;
	*s++ = ' ' ;

	if (reverse)
	    r = print64 (interp, s, p->traf21, 10) ;
	else
	    r = print64 (interp, s, p->traf12, 10) ;
	if (r != TCL_OK)
	    return TCL_ERROR ;

	s = s + strlen (s) ;
	*s++ = ' ' ;

	if (reverse)
	    r = print64 (interp, s, p->traf12, 10) ;
	else
	    r = print64 (interp, s, p->traf21, 10) ;
	if (r != TCL_OK)
	    return TCL_ERROR ;

	Tcl_SetResult (interp, str, TCL_VOLATILE) ;
    }

    return TCL_OK ;
}

/******************************************************************************
Command "<traffichandle> pairs"
******************************************************************************/

int Subtraffic_Pairs (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct traffic *tf ;
    struct pair **elem ;
    char *s ;
    int limit ;

    switch (argc)
    {
	case 2 :
	    limit = -1 ;
	    break ;
	case 3 :
	    if (Tcl_GetInt (interp, argv [2], &limit) != TCL_OK)
		return TCL_ERROR ;
	    break ;
	default :
	    interp->result = "wrong # args: should be traffic pairs ?limit?" ;
	    return TCL_ERROR ;
    }

    tf = (struct traffic *) cd ;

    if (limit <= 0 || limit > tf->nelem)
	limit = tf->nelem ;

    if (tf->nelem == 0)			/* annoying special case */
    {
	interp->result = "" ;
	return TCL_OK ;
    }

    elem = build_array (tf) ;
    if (elem == NULL)
    {
	interp->result = "cannot allocate memory" ;
	return TCL_ERROR ;
    }

    s = write_array_as_list (elem, limit) ;
    if (s == NULL)
    {
	interp->result = "cannot allocate memory" ;
	return TCL_ERROR ;
    }

    free ((char *) elem) ;

    Tcl_SetResult (interp, s, TCL_DYNAMIC) ;
    return TCL_OK ;
}

/******************************************************************************
Command "<traffichandle> sort" and "<traffichandle> sortip"
******************************************************************************/

int comp_traf (const void *v1, const void *v2)
{
    struct pair *p1, *p2 ;
    int64 s1, s2 ;
    int r ;

    p1 = * (struct pair **) v1 ;
    p2 = * (struct pair **) v2 ;

    s1 = p1->traf12 + p1->traf21 ;
    s2 = p2->traf12 + p2->traf21 ;

    if (s1 < s2)
	r = -1 ;
    else if (s1 > s2)
	r = 1 ;
    else r = 0 ;

    return r ;
}

#define	MAX(a,b)	(((a) > (b)) ? (a) : (b))

int comp_ip (const void *v1, const void *v2)
{
    struct pair *p1, *p2 ;
    int r ;
    unsigned long int ip1, ip2 ;

    p1 = * (struct pair **) v1 ;
    p2 = * (struct pair **) v2 ;

    ip1 = MAX (p1->ip1.s_addr, p1->ip2.s_addr) ;
    ip1 = ntohl (ip1) ;
    ip2 = MAX (p2->ip1.s_addr, p2->ip2.s_addr) ;
    ip2 = ntohl (ip2) ;

    if (ip1 < ip2)
	r = 1 ;
    else if (ip1 > ip2)
	r = -1 ;
    else r = 0 ;

    return r ;
}


int Traffic_Sort (ClientData cd, Tcl_Interp *interp, int argc, char *argv [], int (*comp) (const void *, const void*))
{
    struct traffic *tf ;
    struct pair **elem ;
    char *s ;
    int limit ;

    switch (argc)
    {
	case 2 :
	    limit = -1 ;
	    break ;
	case 3 :
	    if (Tcl_GetInt (interp, argv [2], &limit) != TCL_OK)
		return TCL_ERROR ;
	    break ;
	default :
	    interp->result = "wrong # args: should be traffic sort ?limit?" ;
	    return TCL_ERROR ;
    }

    tf = (struct traffic *) cd ;

    if (limit <= 0 || limit > tf->nelem)
	limit = tf->nelem ;

    if (tf->nelem == 0)			/* annoying special case */
    {
	interp->result = "" ;
	return TCL_OK ;
    }

    elem = build_array (tf) ;
    if (elem == NULL)
    {
	interp->result = "cannot allocate memory" ;
	return TCL_ERROR ;
    }

    qsort (elem, tf->nelem, sizeof elem [1], comp) ;

    reverse_array (elem, tf->nelem) ;

    s = write_array_as_list (elem, limit) ;
    if (s == NULL)
    {
	interp->result = "cannot allocate memory" ;
	return TCL_ERROR ;
    }

    free ((char *) elem) ;

    Tcl_SetResult (interp, s, TCL_DYNAMIC) ;
    return TCL_OK ;
}

int Subtraffic_Sort (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return Traffic_Sort (cd, interp, argc, argv, comp_traf) ;
}

int Subtraffic_Sortip (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    return Traffic_Sort (cd, interp, argc, argv, comp_ip) ;
}

/******************************************************************************
Command "<traffichandle> total"
******************************************************************************/

int Subtraffic_Total (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct traffic *tf ;
    struct pair **elem ;
    char str [100] ;
    int64 total ;
    int i, r ;

    if (argc != 2)
    {
	interp->result = "wrong # args: should be traffic total" ;
	return TCL_ERROR ;
    }

    tf = (struct traffic *) cd ;

    elem = build_array (tf) ;
    if (elem == NULL)
    {
	interp->result = "cannot allocate memory" ;
	return TCL_ERROR ;
    }

    total = 0 ;

    for (i = 0 ; i < tf->nelem ; i++)
	total += elem [i]->traf12 + elem [i]->traf21 ;

    free ((char *) elem) ;

    r = print64 (interp, str, total, 10) ;
    if (r == TCL_ERROR)
	return TCL_ERROR ;

    Tcl_SetResult (interp, str, TCL_VOLATILE) ;
    return TCL_OK ;
}

/******************************************************************************
Command "<traffichandle> stats"
******************************************************************************/

int Subtraffic_Stats (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    struct traffic *tf ;
    char str [100] ;
    int i ;
    int used ;
    float eff, qsz ;

    if (argc != 2)
    {
	interp->result = "wrong # args: should be traffic stats" ;
	return TCL_ERROR ;
    }

    tf = (struct traffic *) cd ;

    used = 0 ;
    for (i = 0 ; i < tf->hashsize ; i++)
	if (tf->hashtable [i] != NULL)
	    used ++ ;

    eff = (100.0 * used) / tf->hashsize ;

    qsz = ((float) tf->nelem) / (used == 0 ? 1 : used) ;

    sprintf (str, "%d %d %g %g", tf->hashsize, tf->nelem, eff, qsz) ;
    Tcl_SetResult (interp, str, TCL_VOLATILE) ;

    return TCL_OK ;
}

/******************************************************************************
Command "<traffichandle> delete"
******************************************************************************/

int Subtraffic_Delete (ClientData cd, Tcl_Interp *interp, int argc, char *argv [])
{
    if (argc != 2)
    {
	interp->result = "wrong # args: should be traffic delete" ;
	return TCL_ERROR ;
    }

    Tcl_DeleteCommand (interp, argv [0]) ;

    interp->result = "" ;
    return TCL_OK ;
}
